/*
ArrayOfNetworksGUI
by Cory McKay

CLASS NAME:		ArrayOfNetworksGUI
CLASS TYPE:		Swing GUI
DESCRIPTION:	Provides a general purpose system for creating, training and processing data through
					an array of feedforward neural networks.
				Allows users to specify the number of networks in the array and the number of input nodes,
					number of hidden nodes, number of output nodes, learning rate, momentum factor,
					minimum initial weight and maximum initial weight of each new networks. The initial
					weights of new networks are randomly generated in the given range.
				Allows arrays of networks to be saved and loaded.
				The left display area displays messages regarding the success or failure of actions.
					It can also display the structure of the current array of networks.
				The middle display area can display the currently loaded input data.
				The right display area displays changing errors as networks are trained. It can also
					display the results of processing the currently loaded input data through the
					currently loaded networks.
				The New Network Array button brings up a dialog box allowing the user to specify the
					parameters for creating a new array of networks. The networks are created with
					randomly generated weights when the OK button is pressed. The Define Exception
					button in this dialog box allows the user to specify different settings for the
					network with the specified index. If this is not done, then all networks in the
					array will have the same parameters. The Clear Exceptions button restores
					uniformity if exceptions have been specified.
				The Load Network Array button allows the user to load an array of networks from disk.
					Since a file is used for each network in an array, a directory should be selected
					that contains these files rather than an actual file.
				The Save Network Array button saves the currently loaded network of arrays to a directory.
				The Display Network Array button displays the parameters and state of the currently
					loaded network of arrays in the left display area.
				The Load Input Data button loads a file containing data that can be processed by the
					network or can be used to train the networks.
				The Display Input Data button displays the currently loaded input data in the middle
					display area.
				The Load Model Output button loads model output that should correspond to the currently
					loaded input that can be used to train the networks.
				The Epochs text box specifies the number of training iterations that occur when the
					Train Networks button is pressed.
				The Train Networks button trains the networks using the currently loaded input and
					model output data for the number of iterations specified by the Epochs text box.
					The sum-of-squared error for each 100 iterations is displayed in the right display
					area.
				The Process Data button feeds the currently loaded input data into the networks. The
					output of the networks is displayed in the right display area.
				The sigmoidal activation function is used.
				It should be noted that the sum-of-squared-errors displayed are calculated from the
					actual output from the sigmoidal activation function from the output layer, not
					the scaled values that are displayed.
				All non-integer values are stored and processed as doubles, but are displayed as
					floats for ease of reading.
				This system is designed to need only minor modifications to be used for any task.
					Only the DataIOSystem class will need to be modified, in order to make it compatible
					with the input and expected output files that are used for a particular task.
LIMITATIONS:	Windows-based disk access conventinos were used (namely the use of \ as the directory
					seporator). The load_networks_action and save_networks_action sub-classes of this
					class may need to be modified under other platforms.
				Only one hidden layer is permitted.
*/

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;

class ArrayOfNetworksGUI
	extends JFrame
{
	/* FIELDS ******************************************************************/


	// Array of networks
	NNArray network_array;
	
	// Network I/O data
	double[][][] input_data;			// data sent into network input nodes
	double[][][] output_data;			// model output data
	Object additional_info;				// any additional useful data derived from parsing input_data

	// Dialog boxes
	JFileChooser network_loader;		// used to open network files
	JFileChooser data_loader;			// used to open data files
	JFileChooser saver;					// used to save networks
	NewNetworkArrayDialog new_dialog;	// used to define settings for new networks
	
	// GUI containers
	Container contentPane;
	JPanel network_panel;
	JPanel processing_panel;
	JScrollPane scrollpane_left;
	JScrollPane scrollpane_centre;
	JScrollPane scrollpane_right;
	JTextArea text_area_left;
	JTextArea text_area_centre;
	JTextArea text_area_right;

	// GUI interactive elements related to creating networks
	JButton create_network_button;
	JButton load_network_button;
	JButton save_network_button;
	JButton display_network_button;

	// GUI interactive elements related to processing data through networks
	JButton load_input_data_button;
	JButton display_input_data_button;
	JTextField number_iterations_field;
	JButton train_network_button;
	JButton load_model_output_button;
	JButton run_samples_button;


	/* CONSTRUCTOR *************************************************************/


	ArrayOfNetworksGUI(String title)
	{
		// Initizlize network and data to empty
		network_array = null;
		input_data = null;
		output_data = null;

		// Configure frame and its dialog boxes
		setTitle(title);
		new_dialog = new NewNetworkArrayDialog(ArrayOfNetworksGUI.this);
		network_loader = null;
		data_loader = null;
		saver = null;
		
		// Instantiate containers
		contentPane = getContentPane();
		network_panel = new JPanel();
		processing_panel = new JPanel();
		text_area_left = new JTextArea();
		text_area_centre = new JTextArea();
		text_area_right = new JTextArea();

		// Instantiate interactive elements of GUI related to creating networks
		create_network_button = new JButton("New Network Array");
		load_network_button = new JButton("Load Network Array");
		save_network_button = new JButton("Save Network Array");
		display_network_button = new JButton("Display Network Array");

		// Instantiate interactive elements of GUI related to processing data through networks
		load_input_data_button = new JButton("Load Input Data");
		display_input_data_button = new JButton("Display Input Data");
		load_model_output_button = new JButton("Load Model Output");
		number_iterations_field = new JTextField("5000", 6);
		train_network_button = new JButton("Train Networks");
		run_samples_button = new JButton("Process Data");

		// Prevent inappropriate user actions
		create_network_button.setEnabled(true);
		load_network_button.setEnabled(true);
		save_network_button.setEnabled(false);
		display_network_button.setEnabled(false);
		load_input_data_button.setEnabled(false);
		display_input_data_button.setEnabled(false);
		train_network_button.setEnabled(false);
		load_model_output_button.setEnabled(false);
		run_samples_button.setEnabled(false);

		// Cause buttons to respond to being pressed
		create_network_button.addActionListener(new create_networks_action());
		load_network_button.addActionListener(new load_networks_action());
		save_network_button.addActionListener(new save_networks_action());
		display_network_button.addActionListener(new display_networks_action());
		load_input_data_button.addActionListener(new load_input_data_action());
		display_input_data_button.addActionListener(new display_input_action());
		load_model_output_button.addActionListener(new load_model_output_action());
		train_network_button.addActionListener(new train_network_action());
		run_samples_button.addActionListener(new run_samples_action());

		// Configure the left text display area
		text_area_left.setLineWrap(false);
		text_area_left.setRows(35);
		text_area_left.setColumns(28);
		text_area_left.setEditable(false);
		text_area_left.setText("");
		scrollpane_left = new JScrollPane(text_area_left);

		// Configure the centre text display area
		text_area_centre.setLineWrap(false);
		text_area_centre.setRows(35);
		text_area_centre.setColumns(28);
		text_area_centre.setEditable(false);
		text_area_centre.setText("");
		scrollpane_centre = new JScrollPane(text_area_centre);

		// Configure the right text display area
		text_area_right.setLineWrap(false);
		text_area_right.setRows(35);
		text_area_right.setColumns(28);
		text_area_right.setEditable(false);
		text_area_right.setText("");
		scrollpane_right = new JScrollPane(text_area_right);

		// Configure the GUI components related to creating networks
		network_panel.add(create_network_button);
		network_panel.add(load_network_button);
		network_panel.add(save_network_button);
		network_panel.add(display_network_button);

		// Configure the GUI components related to processing data in the network
		processing_panel.add(load_input_data_button);
		processing_panel.add(display_input_data_button);
		processing_panel.add(load_model_output_button);
		processing_panel.add(new JLabel("Epochs:"));
		processing_panel.add(number_iterations_field);
		processing_panel.add(train_network_button);
		processing_panel.add(run_samples_button);

		// Caus all GUI components to be displayed
		contentPane.add(network_panel, BorderLayout.NORTH);
		contentPane.add(scrollpane_left, BorderLayout.WEST);
		contentPane.add(scrollpane_centre, BorderLayout.CENTER);
		contentPane.add(scrollpane_right, BorderLayout.EAST);
		contentPane.add(processing_panel, BorderLayout.SOUTH);

		// Cause program to quit when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				System.exit(0);
			}
		});
	}


	/* SUB-CLASSES *************************************************************/


	class create_networks_action extends AbstractAction
	// Display a dialog box allowing the user to set the configurations of a new array
	// of neural networks. If the user presses OK, a new NNArray is created, replacing
	// any existing one in memory. Its weights are randomly generated. No change occurs
	// if the cancel butotn is pressed.
	{
		public void actionPerformed(ActionEvent event)
		{
			new_dialog.show();
			if (new_dialog.was_updated() == true) // take action if the OK button was pressed
			{
				network_array = new_dialog.generate_new_network_array();

				// Update what user is permitted to do
				save_network_button.setEnabled(true);
				display_network_button.setEnabled(true);
				load_input_data_button.setEnabled(true);
				display_input_data_button.setEnabled(false);
				train_network_button.setEnabled(false);
				load_model_output_button.setEnabled(false);
				run_samples_button.setEnabled(false);

				// Display action succesful message
				text_area_left.setText("Array of " + network_array.length + " networks " +
					"succesfully created");
				text_area_centre.setText("");
				text_area_right.setText("");
			}
		}
	}


	class load_networks_action extends AbstractAction
	// Bring up a JFileChooser dialog box to load an array of pre-trained networks.
	{
		public void actionPerformed(ActionEvent event)
		{
			if (network_loader == null) // create a new JFileChooser if one doesn't exist already
			{
				network_loader = new JFileChooser();
				network_loader.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				network_loader.setCurrentDirectory(new File("."));
			}

			int loader_return = network_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				File selection = network_loader.getSelectedFile();
				NetworksFileReader reader = new NetworksFileReader();
				try
				{
					// Determine number of networks there are by reading guide.dat
					int number_networks = reader.get_number_networks(selection.getPath() + "\\guide.dat");

					// Prepare data stuctures to hold loaded data
					NetworkSettings[] network_info = new NetworkSettings[number_networks];
					double[][] weights = new double[number_networks][];

					// Process network files
					for (int i = 0; i < number_networks; i++)
					{
						// Load data from file into network_data
						reader = new NetworksFileReader();
						reader.read_network_file(selection.getPath() + "\\nn" + i + ".dat");
						double[] network_data = reader.get_network_data(); // the contents of the network file
						
						// Load settings and weights
						network_info[i] = new NetworkSettings(network_data[0], network_data[1],
							(int) network_data[2], (int) network_data[3], (int) network_data[4], 0, 0);
						weights[i] = new double[network_data.length - 5];
						for (int j = 0; j < weights[i].length; j++)
							weights[i][j] = network_data[ j + 5 ];
					}

					// Set up array
					network_array = new NNArray(network_info, weights);

					// Update what user is permitted to do
					save_network_button.setEnabled(true);
					display_network_button.setEnabled(true);
					load_input_data_button.setEnabled(true);
					display_input_data_button.setEnabled(false);
					train_network_button.setEnabled(false);
					load_model_output_button.setEnabled(false);
					run_samples_button.setEnabled(false);

					// Report success
					text_area_left.setText("Array of " + network_array.length + " networks " +
						"succesfully created");
					text_area_centre.setText("");
					text_area_right.setText("");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Inappropriate directory specified.\n");
					text_area_centre.setText("");
					text_area_right.setText("");
				}				
			}			
		}
	}

	class save_networks_action extends AbstractAction
	// Brings up a JFileChooser dialog box to save the currently loaded array of networks
	// to disk. Only directories may be selected, as a file will be created for each
	// network in the array. If a directory does not already exist, it will be created.
	// A file named guide.dat is created containing the number of networks saved. The
	// networks are saved in files named nni.dat, where i is a number from 0 to number
	// of networks - 1.
	{
		public void actionPerformed(ActionEvent event)
		{
			if (saver == null) // create a new JFileChooser if one doesn't exist already
			{
				saver = new JFileChooser();
				saver.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				saver.setCurrentDirectory(new File("."));
			}

			int saver_return = saver.showSaveDialog(ArrayOfNetworksGUI.this);
			if (saver_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				File selection = saver.getSelectedFile();
				FWriter writer = new FWriter();
				try
				{
					// Create directory if does not exist already
					if (selection.isFile() == false && selection.isDirectory() == false)
						selection.mkdir();

					// Write files
					writer.write_file(new String(selection.getPath() + "\\guide.dat"),
					                  String.valueOf(network_array.length));
					for (int i = 0; i < network_array.length; i++)
					{
						writer.write_file(new String(selection.getPath() + "\\nn" + i + ".dat"),
										  network_array.get_controller(i).get_network_data());
					}

					// Display success message
					text_area_left.setText("Networks saved succesfully in:\n" +
					                       selection.getPath() + "\n");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Unable to save networks\n");
				}				
			}	
		}
	}


	class display_networks_action extends AbstractAction
	// Display the structure of the network in the left display area
	{
		public void actionPerformed(ActionEvent event)
		{
			String[] network_contents = network_array.expose_networks();
			NetworkSettings[] network_settings = network_array.get_network_data();
			
			String output_text = new String("NETWORK STRUCTURE\n\n\n");
			for (int i = 0; i < network_contents.length; i++)
			{
				output_text += "NETWORK " + i + "\n\n";
				output_text += "Learning Rate: " + network_settings[i].learning_rate + "\n";
				output_text += "Momentum: " + network_settings[i].momentum + "\n";
				output_text += "Input Nodes: " + network_settings[i].num_input_nodes + "\n";
				output_text += "Hidden Nodes: " + network_settings[i].num_hidden_nodes + "\n";
				output_text += "Output Nodes: " + network_settings[i].num_output_nodes + "\n\n";
				output_text += network_contents[i];
				output_text += "\n\n\n";
			}

			text_area_left.setText(output_text);
		}
	}


	class load_input_data_action extends AbstractAction
	// Bring up a JFileChooser dialog box to open a data file containing data to
	// be run through the network. This data may be used for training or may
	// just be data to be processed.
	{
		public void actionPerformed(ActionEvent event)
		{
			int loader_return;
			if (data_loader == null) // create a new JFileChooser if one doesn't exist already
			{
				data_loader = new JFileChooser();
				data_loader.setCurrentDirectory(new File("."));
			}

			loader_return = data_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				DataIOSystem reader = new DataIOSystem();
				try
				{
					// Read input data into input_data array
					input_data = reader.get_input_data(data_loader.getSelectedFile().getPath(),
					                                   network_array.length,
					                                   network_array.get_network_data());

					// Get any additional useful data from the input data
					additional_info = reader.get_additional_input_info(data_loader.getSelectedFile().getPath(),
					                                                   network_array.length,
					                                                   network_array.get_network_data());

					// Update what user is permitted to do
					display_input_data_button.setEnabled(true);
					run_samples_button.setEnabled(true);
					load_model_output_button.setEnabled(true);
					
					// Display success message
					text_area_left.setText("Input data loaded succesfully.\n");
					text_area_centre.setText("");
					text_area_right.setText("");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Inappropriate input data file specified.\n");
				}				
			}
		}
	}


	class display_input_action extends AbstractAction
	// Display the loaded input data in the middle display area
	{
		public void actionPerformed(ActionEvent event)
		{
			String print_string = new String("LOADED INPUT DATA");
			for (int i = 0; i < input_data.length; i++)
			{
				print_string += "\n\n\nNEURAL NETWORK " + i + "\n\n";
				for (int j = 0; j < input_data[i].length; j++)
				{
					print_string += "Network " + i + "   Input Set " + j + "\n";
					for (int k = 0; k < input_data[i][j].length; k++)
					{
						print_string += "    Input Node: " + k + "   Value: " +
										input_data[i][j][k] + "\n";
					}
				}
			}
			text_area_centre.setText(print_string);
		}
	}


	class load_model_output_action extends AbstractAction
	// Bring up a JFileChooser dialog box to open a data file containing model output
	// data that the networks can be trained on.
	{
		public void actionPerformed(ActionEvent event)
		{
			int loader_return;
			if (data_loader == null) // create a new JFileChooser if one doesn't exist already
			{
				data_loader = new JFileChooser();
				data_loader.setCurrentDirectory(new File("."));
			}

			loader_return = data_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				DataIOSystem reader = new DataIOSystem();
				try
				{
					// Read output data into output_data array
					output_data = reader.get_model_output_data(data_loader.getSelectedFile().getPath(),
													           network_array.length,
													           network_array.get_network_data(),
					                                           additional_info);

					// Update what user is permitted to do
					train_network_button.setEnabled(true);

					// Display success message
					text_area_left.setText("Model output data loaded succesfully.\n");
				}
				catch (Exception e)
				{
					text_area_left.setText("ERROR:\n" + "Inappropriate model output data file specified.\n");
				}				
			}
		}
	}


	class train_network_action extends AbstractAction
	// Trains the networks using the loaded input data and expected output data.
	// Displays the sum-of-squared-errors for every 100th epoch in the right display area.
	{
		public void actionPerformed(ActionEvent event)
		{
			// Train network and produce array of the average sum-of-squared-error for
			// each training iteration of each network
			double[][] error_values = network_array.train_networks
			(
				Integer.parseInt(number_iterations_field.getText()),
				input_data,
				output_data
			);

			// Print average sum-of-squared error for every 100th epoch of each network
			Double val; // used for rounding for display purposes
			String error_list = new String("AVERAGE SUM-OF-SQUARED ERRORS\n"); // used to display errors
			for (int i = 0; i < error_values.length; i++)
			{
				error_list += "\n\nNETWORK " + i + "\n";
				for (int j = 0; j < error_values[i].length; j++)
				{
					if (j % 100 == 0)
					{
						val = new Double(error_values[i][j]);
						error_list += "\n   Network: " + i + "   Epoch: " + j + "   Av. Error: " +
						              val.floatValue() + "\n";
					}
				}
			}
			text_area_right.setText(error_list);
		}
	}


	class run_samples_action extends AbstractAction
	// Displays the results of passing the loaded input data through the network
	// in the central display area. The results that are displayed are in the
	// form usable to the user, not the raw form stored in output_data.
	{
		public void actionPerformed(ActionEvent event)
		{
			// Get unformatted array of outputs
			double[][][] results = network_array.get_outputs(input_data);
			
			// Translate results
			DataIOSystem interpreter = new DataIOSystem();
			double[][][] interpreted_results = interpreter.interpret_results(results);

			// Print translated results in right display area
			Double val; // used for rounding for display purposes
			String output_list = new String("OUTPUT OF NETWORKS\n"); // used to display output
			for (int i = 0; i < interpreted_results.length; i++)
			{
				output_list += "\n\nNEURAL NETWORK " + i + "\n\n";
				for (int j = 0; j < interpreted_results[i].length; j++)
				{
					output_list += "Network " + i + "   Input Set " + j + "\n";
					for (int k = 0; k < interpreted_results[i][j].length; k++)
					{
						val = new Double(interpreted_results[i][j][k]);
						output_list += "    Output Node: " + k + "   Value: " +
										 val.floatValue() + "\n";
					}
				}
			}
			text_area_right.setText(output_list);
		}
	}
}