/*
DataIOSystem
by Cory McKay

CLASS NAME:		NetworksFileReader
CLASS TYPE:		File reader and processor
DESCRIPTION:	Reads in input files and expected output files that can be processed through
				neural networks and can be used to train the networks. Although the other classes
				of this system are designed to be multipurpose, the following methods of this
				class may need to be modified to meet the format of input and output files
				that will be used for particular applicatoins:
					get_additional_input_info
					interpret_results
					parse_input_data
					parse_model_output_data
					scale_input_data
					scale_model_output_data
				Objects of this class also translate raw network output data into a meaningufl
				form (using the interpret_results method).
				The get_input_data method can be called to a load a file containing input data
				for a network. This data can be used to train networks or can be just processed
				by the network.
				The get_additional_input_info method can be called after the get_input_data to get
				any additional useful information after the input data has been called. This will
				not always be necessary, in which case this method should be set to return null.
				The get_model_output_data method can be called to load a file containg model output
				that corresponds to the loaded input data. This is used for training the networks.
*/

import java.io.*;
import java.util.StringTokenizer;

class DataIOSystem
{
	/* FIELDS ******************************************************************/


	private int next_index;				// the current location in the chars array
	private char[] chars;				// an array that holds all of the characters that are read in through
										//     the file


	/* PUBLIC METHODS **********************************************************/


	public double[][][] get_input_data(String filename, int number_networks, NetworkSettings[] network_data)
		throws IOException
	// Parses the given file to correspond with the given network configurations.
	// Extracts input data from the file that can be fed to an array of networks.
	// Scales the results so that all inputs fall between 0.25 and 0.75.
	// The first index of the retruned array specifies the network index, the
	// second specifies data set number and the third specifies input node.
	{
		// Initialize fields
		chars = null;
		next_index = 0;

		// Read input data file into chars
		access_file(filename);

		// Parse the contents of the file
		double[][][] input_data = parse_input_data(number_networks, network_data);

		// Scale input data to 0.25 to 0.75 range (sigmoidal activation function)
		scale_input_data(input_data, number_networks, network_data);
		
		// Reset fields
		chars = null;
		next_index = 0;

		// Return scaled input data
		return input_data;
	}


	public Object get_additional_input_info(String filename, int number_networks,
	                                        NetworkSettings[] network_data)
		throws IOException
	// Return any additional information from the input data file that might be useful
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		// This is not necessary in this case
		return null;
	}


	public double[][][] get_model_output_data(String filename, int number_networks,
	                                          NetworkSettings[] network_data, Object additional_info)
		throws IOException
	// Parses the given file to correspond with the given network configurations.
	// Extracts model output data from the file that can be used to train an array of networks.
	// Scales the results so that all inputs fall between 0.25 and 0.75.
	// The first index of the retruned array specifies the network index, the
	// second specifies data set number and the third specifies input node.
	{
		// Initialize fields
		chars = null;
		next_index = 0;

		// Read input data file into chars
		access_file(filename);

		// Parse the contents of the file
		double[][][] model_output_data = parse_model_output_data(number_networks, network_data,
		                                                         additional_info);

		// Scale output data to 0.25 to 0.75 range (sigmoidal activation function)
		scale_model_output_data(model_output_data, number_networks, network_data, additional_info);
		
		// Reset fields
		chars = null;
		next_index = 0;

		// Return scaled input data
		return model_output_data;
	}


	public double[][][] interpret_results(double[][][] raw_results)
	// Returns an array holding a translation of the raw output from network into a form
	// useful for the application at hand.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		double[][][] translation = new double[raw_results.length][][]; // translated results
		
		// In this case, set every result below 0.5 to 0 and and every result above it to 1
		for (int i = 0; i < raw_results.length; i++)
		{
			translation[i] = new double[raw_results[i].length][];
			for (int j = 0; j < raw_results[i].length; j++)
			{
				translation[i][j] = new double[raw_results[i][j].length];
				for (int k = 0; k < raw_results[i][j].length; k++)
				{
					if (raw_results[i][j][k] < 0.5) translation[i][j][k] = 0;
					else translation[i][j][k] = 1;
				}
			}
		}

		return translation;
	}


	/* PRIVATE METHODS *********************************************************/

	
	private void access_file(String filename) 
		throws IOException
	// Returns the contents of the specified file in the chars array
	{
		File src = new File(filename);
		chars = new char[(int) src.length()];
		src = null;
		
		InputStream in;
		in = new FileInputStream(filename);
		int ch;
		for (int i = 0; (ch = in.read()) != -1; i++)
			chars[i] = (char)ch;
		in.close();
	}


	private String get_next_string()
	// Reads in one line from chars and returns it as a string. Also moves next_index to
	// the next_location in the file.
	{
		int i = next_index;
		int done = -1;

		while (done != 1)
		{
			if (i < chars.length && done != 1)
			{
				if ( chars[i] != '\n' )
					i++;
				else done = 1;
			}
			else done = 1;
		}
		String temp = new String(chars, next_index, i-next_index);
		next_index = i + 1;
		return temp;
	}


	private String find_line_starting_with(String search_string)
	// Return the first line from the current location in chars that starts with search_string
	{
		String current_line = null; // the current line of information being read
		StringTokenizer tokens; // used to parse each line of information
		boolean done = false; // keeps track of when loop is finished
		while (done == false)
		{
			current_line = get_next_string();		
			tokens = new StringTokenizer(current_line, " ");
			if (tokens.nextToken().equals(search_string))
				done = true;
		}
		return current_line;
	}


	private double[][][] parse_input_data(int number_networks, NetworkSettings[] network_data)
	// Parses the data that has been read into chars and that corresponds to the given
	// network. Returns it in as an array whose first index specifies the network index,
	// second specifies data set number and the third specifies input node.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		// In this case, reads in data from a file such as logicTrainingData.txt
		// Each network corresponds to a boolean logic function. Each training
		// set is on a new line. The training data for each input is followed
		// by the model output data for each output (in the case of
		// logicTrainingData.txt, the first output corresponds to the function's
		// and the second corresponds to the logical opposite. Although the
		// training data and model output data are both contained in the same
		// file in this example, this is not necessary.
		double[][][] data = new double[number_networks][][]; // data to be returned
		String current_line = null; // the current line of information being read
		StringTokenizer tokens; // used to parse each line of information

		for (int i = 0; i < number_networks; i ++)
		{
			data[i] = new double[network_data[i].num_input_nodes * network_data[i].num_input_nodes][];
			current_line = get_next_string(); // skip irrelevant line
			for (int j = 0; j < network_data[i].num_input_nodes * network_data[i].num_input_nodes; j++)
			{
				data[i][j] = new double[network_data[i].num_input_nodes];
				current_line = get_next_string();
				tokens = new StringTokenizer(current_line, " ");

				for (int k = 0; k < network_data[i].num_input_nodes; k++)
					data[i][j][k] = Double.parseDouble(tokens.nextToken());
			}
			current_line = get_next_string(); // skip blank line
		}

		return data;
	}


	private double[][][] parse_model_output_data(int number_networks, NetworkSettings[] network_data,
	                                             Object additional_info)
	// Parses the data that has been read into chars and that corresponds to the biven
	// network. Returns it in as an array whose first index specifies the network index,
	// second specifies data set number and the third specifies input node.
	// additional_info corresponds to any extra data that is needed to parse the model
	// output.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		// In this case, reads in data from a file such as logicTrainingData.txt
		// Each network corresponds to a boolean logic function. Each training
		// set is on a new line. The training data for each input is followed
		// by the model output data for each output (in the case of
		// logicTrainingData.txt, the first output corresponds to the function's
		// and the second corresponds to the logical opposite. Although the
		// training data and model output data are both contained in the same
		// file in this example, this is not necessary.
		double[][][] data = new double[number_networks][][]; // data to be returned
		String current_line = null; // the current line of information being read
		StringTokenizer tokens; // used to parse each line of information

		for (int i = 0; i < number_networks; i ++)
		{
			data[i] = new double[network_data[i].num_input_nodes * network_data[i].num_input_nodes][];
			current_line = get_next_string(); // skip irrelevant line
			for (int j = 0; j < Math.pow(network_data[i].num_input_nodes, 2); j++)
			{
				data[i][j] = new double[network_data[i].num_input_nodes];
				current_line = get_next_string();
				tokens = new StringTokenizer(current_line, " ");

				String temp; // dummy string
				for (int k = 0; k < network_data[i].num_input_nodes; k++)
					temp = tokens.nextToken(); // skip input data
				for (int k = 0; k < network_data[i].num_output_nodes; k++)
					data[i][j][k] = Double.parseDouble(tokens.nextToken());
			}
			current_line = get_next_string(); // skip blank line
		}

		return data;
	}


	private void scale_input_data(double[][][] data, int number_networks,
	                              NetworkSettings[] network_data)
	// Scale the input data so that it is of a convenient size to serve as input to
	// the networks.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		for (int i = 0; i < data.length; i++)
		{
			for (int j = 0; j < data[i].length; j++)
			{
				for (int k = 0; k < data[i][j].length; k++)
					data[i][j][k] = (data[i][j][k] * 0.5) + 0.25; // scale to between 0.25 and 0.75
			}
		}
	}


	private void scale_model_output_data(double[][][] data, int number_networks,
	                                     NetworkSettings[] network_data, Object additional_info)
	// Scale the model outpu data so that it falls within the ranges of 0.25 and 0.75.
	// This is so that it can be readily achieved by sigmoidal activation functions.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		for (int i = 0; i < data.length; i++)
		{
			for (int j = 0; j < data[i].length; j++)
			{
				for (int k = 0; k < data[i][j].length; k++)
					data[i][j][k] = (data[i][j][k] * 0.5) + 0.25; // scale to between 0.25 and 0.75
			}
		}
	}
}