/*
NNArray
by Cory McKay

CLASS NAME:		NNArray
CLASS TYPE:		Array of neural network controllers
DESCRIPTION:	Objects of this class hold an array of NNControllers, each with its own network
				and network settings.
				Adjustable learning rates and momentums are permitted.
				An arbitrary number of input nodes, hidden nodes and output nodes are permitted.
LIMITATIONS:	Only applicable to perceptrons or neural networks with one layer of hidden nodes.
				Only limited error checking is implemented. It is assumed that correct input is passed to the methods.
				Nodes uses sigmoidal activation function, so all output and expected output should be scaled when received
				from and sent to controllers of this class so that it falls between 0.25 and 0.75.
*/

class NNArray
{
	/* FIELDS ******************************************************************/


	private NNController[] networks;	// array of neural network controllers
	private NetworkSettings[] settings;	// settings for the parameters used by the networks controllers
	public int length;					// number of networks contained in NNArray


	/* CONSTRUCTORS ************************************************************/


	public NNArray(NetworkSettings[] info)
	// Create an array of new, untrained neural networks based on the given settings.
	{
		settings = info;
		networks = new NNController[settings.length];
		for (int i = 0; i < settings.length; i++)
		{
			networks[i] = new NNController();
			networks[i].create_network(settings[i].learning_rate, settings[i].momentum,
				settings[i].num_input_nodes, settings[i].num_hidden_nodes, settings[i].num_output_nodes,
				settings[i].min_initial_weight, settings[i].max_initial_weight);
		}
		length = settings.length;
	}


	public NNArray(NetworkSettings[] info, double[][] weights)
	// Create an array pre-trained neural networks based on the given settings. The first
	// index of the weights array specifies the network that it refers to. The rest of the
	// weights array contains the weights of the hidden nodes (if any) followed by those of
	// the output nodes.
	{
		settings = info;
		networks = new NNController[settings.length];
		for (int i = 0; i < settings.length; i++)
		{
			networks[i] = new NNController();
			networks[i].load_network(settings[i].learning_rate, settings[i].momentum,
				settings[i].num_input_nodes, settings[i].num_hidden_nodes, settings[i].num_output_nodes,
				weights[i]);
		}
		length = settings.length;
	}


	/* PUBLIC METHODS **********************************************************/


	public double[][] train_networks(int iterations, double[][][] training_input, double[][][] expected_output)
	// Train the networks using the given array of possible sets of inputs and the given array
	// of corresponding sets of desired outputs for each network. The first index of training_input
	// and of expended_output specifies the network.
	// The first index of the returned array specifies the network that the returned data corresponds
	// to. The portion of the return array for each network holds the average sum-of-squared-error for
	// each training iteration.
	{
		double[][] average_error = new double[networks.length][];  // the average sum-of-squared-error for
		                                                           // each training iteration for each network
		for (int i = 0; i < networks.length; i++)
			average_error[i] = networks[i].train_network(iterations, training_input[i], expected_output[i]);
		return average_error;
	}


	public double[][][] get_outputs(double[][][] input_values)
	// Return the output of each network for the given set of inputs.
	// The first index of input_values specifies the network, the second specifis the data set and
	// the third specifies the input node.
	// The first index of the returned array specifies the network, the second specifis the data set and
	// the third specifies the output node.
	{
		double[][][] output = new double[networks.length][][];
		for (int i = 0; i < input_values.length; i++)
		{
			output[i] = new double[input_values[i].length][];
			for (int j = 0; j < input_values[i].length; j++)
				output[i][j] = networks[i].get_output(input_values[i][j]);
		}
		return output;
	}


	public double[] get_errors(double[][][] given_input, double[][][] expected_output)
	// Returns the error for each network corresponding to the given data.
	// The first index of given_input and of expended_output specifies the network.
	{
		double error[] = new double[networks.length];
		for (int i = 0; i < networks.length; i++)
			error[i] = networks[i].get_error(given_input[i], expected_output[i]);
		return error;
	}


	public NNController get_controller(int index)
	// Get the NNController in networks specified by the given index
	{
		return networks[index];
	}


	public String[] expose_networks()
	// Returns an array of strings containing the current inputs, weights and values of the
	// network nodes for each network. Used during debugging.
	{
		String[] configuration = new String[networks.length];
		for (int i = 0; i < networks.length; i++)
			configuration[i] = networks[i].expose_network();
		return configuration;
	}


	public NetworkSettings[] get_network_data()
	// Returns the settings for the networks.
	{
		return settings;
	}


	public NNController get_NNController(int i)
	// Returns the NNController at the given location in networks.
	{
		return networks[i];
	}
}